#ifndef CMD_PROCESSOR_H
#define CMD_PROCESSOR_H

#include <iostream>
#include <iomanip>
#include <sstream>
#include <limits.h>
#include <cstring>
#include <iterator> 
#include <chrono>

#include <gbs_mca_comm.h>
#include <gbs_mca_spefile.h>
#include <gbs_mca_data.h>

using namespace std;

// Constants -------------------------------------------------------------------------------

const string SpeDefaultName = "mca_spectrum.spe";

// Generic error which provides a message
struct MsgError  //--------------------------------------------------------------------------
{
    string msg;
    MsgError(string m = string()) { msg = m; }
};


class Cmd_Processor //----------------------------------------------------------------------
{
protected:

    struct McaComm::GBS_MCA_Comm *mca;
    struct McaComm::QUERY_STATE *mcaState;
    struct McaComm::QUERY_STATE527 *mcaState527;
    struct McaComm::QUERY_STATE527_EX *mcaState527_ex;
    struct McaComm::QUERY_POWER *mcaPower;
    struct McaComm::QUERY_SYSTEM_DATA *mcaSystem;
    McaData::GBS_MCA_Data *mcaData;

    uint16_t MaxHv;
    uint16_t MaxChannels;
    uint16_t MaxUld;
    uint16_t MaxFlattop;    // * 0.1 μs
    bool verbose;


private:
    vector<string> split(string input);
    void DoSave(string filename, string comment);
    void DoSetHv(uint16_t hv, int32_t inhibit);
    void DoSetPreset(std::string preset, uint32_t time);
    void DoSetAdc(uint16_t channels, uint16_t lld, uint16_t uld);
    void DoSetThreshold(float threshold);
    void DoSetGain(uint16_t coarseGain, uint16_t fineGain);
    void DoSetAmpPolarity(bool positive);
    void DoSetPzc(uint16_t pzc_value);
    void DoAutoPzc();
    void DoSetShapingTime(float shaping_time);
    void DoSetTriggerFilter(uint16_t index);
    void DoSetFlattopTime(float ft_time);
    void DoReset();
    void DoQueryState();

public:
    Cmd_Processor();
    ~Cmd_Processor();

    bool init(int argc, char *argv[]);
    bool parse(string line);
};

#endif // CMD_PROCESSOR_H
